/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  3-5-10                                                       */
/*                                                               */
/* 8/24/10  - fixed memory leak in AddSequenceCompleteBasis      */
/*             by freeing the passed-in sequence, version 1.1    */
/*****************************************************************/

#define _CLUSTERADT

/*********************************************************************************/
/* Included Header Files */
/*********************************************************************************/
#include "Globals.h"
#include "MatrixUlongADT.h"
#include "MatrixDoubleADT.h"
#include "StringADT.h"
#include "SuffixTreeADT.h"
#include "SuffixTreeListADT.h"
#include "ClusterADT.h"
#include "Gram.h"

/*********************************************************************************/
/* Private Module Constants */
/*********************************************************************************/
#define CLUSTERADT_CHUNK_SIZE  10

/*********************************************************************************/
/* Private Type Definitions */
/*********************************************************************************/

/*********************************************************************************/
/* Private Module Function Prototypes */
/*********************************************************************************/

/*********************************************************************************/
/* Private Module Variables */
/*********************************************************************************/

/*********************************************************************************/
/* Public and Private Module Functions */
/*********************************************************************************/

void CLUSTERADT_InitializeCluster(CLUSTERADT_CLUSTER_TYPE *Cluster, 
                                    STADT_TREE_TYPE *Sequence,
                                    ULONG SequenceIndex,
                                    FLOAT DictionarySize,
                                    UBYTE MaxNumberBasisSequences)
{
   Cluster->NumberOfBasesAllocated = MaxNumberBasisSequences;
   Cluster->NumberOfSequencesAllocated = CLUSTERADT_CHUNK_SIZE;

   Cluster->MarkedAsCompleteBasis = FALSE;
   
   Cluster->NumberOfBases = 1;
   Cluster->NumberOfSequences = 1;

   MATRIXULONGADT_InitializeMatrix(&(Cluster->SequenceIndicies), 1, Cluster->NumberOfSequencesAllocated);
   MATRIXDOUBLEADT_InitializeMatrix(&(Cluster->SequenceDictionaries), 1, Cluster->NumberOfSequencesAllocated);
   MATRIXDOUBLEADT_InitializeMatrix(&(Cluster->DistanceVectors), Cluster->NumberOfSequencesAllocated, MaxNumberBasisSequences);

   MATRIXULONGADT_InitializeMatrix(&(Cluster->BasisSequenceIndicies), 1, MaxNumberBasisSequences);

   Cluster->RepresentationSequence = NULL;
   
   Cluster->SequenceIndicies.Rows[0].Cols[0] = SequenceIndex;
   Cluster->BasisSequenceIndicies.Rows[0].Cols[0] = 0;

   STLADT_InitializeTreeList(&(Cluster->Sequences));
   STLADT_AppendTreeToTreeList(&(Cluster->Sequences), Sequence);

   Cluster->SequenceDictionaries.Rows[0].Cols[0] = DictionarySize;

   Cluster->DistanceVectors.Rows[0].Cols[0] = 0;

   Cluster->RepresentationIndex = 0;
}

/*********************************************************************************/

void CLUSTERADT_CleanUpCluster(CLUSTERADT_CLUSTER_TYPE *Cluster)
{
   Cluster->NumberOfBasesAllocated = 0;
   Cluster->NumberOfSequencesAllocated = 0;

   Cluster->MarkedAsCompleteBasis = FALSE;

   Cluster->NumberOfBases = 0;
   Cluster->NumberOfSequences = 0;

   MATRIXULONGADT_CleanUpMatrix(&(Cluster->SequenceIndicies));
   MATRIXDOUBLEADT_CleanUpMatrix(&(Cluster->SequenceDictionaries));
   MATRIXDOUBLEADT_CleanUpMatrix(&(Cluster->DistanceVectors));

   MATRIXULONGADT_CleanUpMatrix(&(Cluster->BasisSequenceIndicies));

   STLADT_CleanUpTreeList(&(Cluster->Sequences));
   
   if (Cluster->RepresentationSequence != NULL)
   {
      STADT_DeleteTree (Cluster->RepresentationSequence);
      Cluster->RepresentationSequence = NULL;
   }
}

/*********************************************************************************/

void CLUSTERADT_AddSequenceIncompleteBasis(CLUSTERADT_CLUSTER_TYPE *Cluster, 
                                             STADT_TREE_TYPE *Sequence,
                                             ULONG SequenceIndex,
                                             FLOAT DictionarySize,
                                             BOOLEAN AddNewBasis)
{
   ULONG j;
   FLOAT Distance;
   
   Cluster->NumberOfSequences += 1;

   if (Cluster->NumberOfSequences >= Cluster->NumberOfSequencesAllocated)
   {
      Cluster->NumberOfSequencesAllocated += CLUSTERADT_CHUNK_SIZE;
      
      MATRIXULONGADT_RedefineMatrix(&(Cluster->SequenceIndicies), 1, Cluster->NumberOfSequencesAllocated);
      MATRIXDOUBLEADT_RedefineMatrix(&(Cluster->SequenceDictionaries), 1, Cluster->NumberOfSequencesAllocated);
      MATRIXDOUBLEADT_RedefineMatrix(&(Cluster->DistanceVectors), Cluster->NumberOfSequencesAllocated, Cluster->NumberOfBasesAllocated);
   }

   Cluster->SequenceIndicies.Rows[0].Cols[Cluster->NumberOfSequences - 1] = SequenceIndex;

   STLADT_AppendTreeToTreeList(&(Cluster->Sequences), Sequence);

   Cluster->SequenceDictionaries.Rows[0].Cols[Cluster->NumberOfSequences - 1] = DictionarySize;

   for (j = 0; j < Cluster->NumberOfBases; j++)
   {
      Distance = GRAM_CalculateDistance(Cluster->Sequences.Trees[Cluster->BasisSequenceIndicies.Rows[0].Cols[j]], 
                                       Cluster->SequenceDictionaries.Rows[0].Cols[Cluster->BasisSequenceIndicies.Rows[0].Cols[j]],
                                       Sequence,
                                       DictionarySize);

      Cluster->DistanceVectors.Rows[Cluster->NumberOfSequences - 1].Cols[j] = Distance;
   }
   
   if (AddNewBasis == TRUE)
   {
      Cluster->NumberOfBases += 1;
      Cluster->BasisSequenceIndicies.Rows[0].Cols[Cluster->NumberOfBases - 1] = Cluster->NumberOfSequences - 1;

      for (j = 0; j < Cluster->NumberOfSequences; j++)
      {
         Distance = GRAM_CalculateDistance(Cluster->Sequences.Trees[j], 
                                          Cluster->SequenceDictionaries.Rows[0].Cols[j],
                                          Sequence,
                                          DictionarySize);
   
         Cluster->DistanceVectors.Rows[j].Cols[Cluster->NumberOfBases - 1] = Distance;
      }
   }
}

/*********************************************************************************/

void CLUSTERADT_AddSequenceCompleteBasis(CLUSTERADT_CLUSTER_TYPE *Cluster, 
                                          STADT_TREE_TYPE *Sequence,
                                          ULONG SequenceIndex,
                                          FLOAT DictionarySize)
{
   Cluster->NumberOfSequences += 1;

   if (Cluster->NumberOfSequences >= Cluster->NumberOfSequencesAllocated)
   {
      Cluster->NumberOfSequencesAllocated += CLUSTERADT_CHUNK_SIZE;
      
      MATRIXULONGADT_RedefineMatrix(&(Cluster->SequenceIndicies), 1, Cluster->NumberOfSequencesAllocated);
   }
   
   Cluster->SequenceIndicies.Rows[0].Cols[Cluster->NumberOfSequences - 1] = SequenceIndex;

   if (Sequence != NULL)
   {
      STADT_DeleteTree (Sequence);
   }
}

/*********************************************************************************/

void CLUSTERADT_SetClusterAsCompleteBasis(CLUSTERADT_CLUSTER_TYPE *Cluster)
{
   Cluster->MarkedAsCompleteBasis = TRUE;

   Cluster->RepresentationSequence = Cluster->Sequences.Trees[Cluster->RepresentationIndex];
   Cluster->Sequences.Trees[Cluster->RepresentationIndex] = NULL;
   Cluster->RepresentationDictionary = Cluster->SequenceDictionaries.Rows[0].Cols[Cluster->RepresentationIndex];

   MATRIXDOUBLEADT_CleanUpMatrix(&(Cluster->SequenceDictionaries));
   MATRIXDOUBLEADT_CleanUpMatrix(&(Cluster->DistanceVectors));

   MATRIXULONGADT_CleanUpMatrix(&(Cluster->BasisSequenceIndicies));

   STLADT_CleanUpTreeList(&(Cluster->Sequences));
}
