/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  3-5-10                                                       */
/*                                                               */
/* 9/27/10  - Added reverse complement checking, version 1.3     */
/*****************************************************************/

#define _FILEIO

/*********************************************************************************/
/* Included Header Files */
/*********************************************************************************/
#include "Globals.h"
#include "MatrixUlongADT.h"
#include "MatrixDoubleADT.h"
#include "StringADT.h"
#include "SuffixTreeADT.h"
#include "SuffixTreeListADT.h"
#include "StringListADT.h"
#include "ClusterADT.h"
#include "ClusterListADT.h"
#include "FileIO.h"
#include "Parameters.h"

/*********************************************************************************/
/* Private Module Constants */
/*********************************************************************************/
#define FILEIO_SEQUENCE_PART_LENGTH                   512
#define FILEIO_BASE(a)                                (((m_ReverseSequence == TRUE) && (a == 'A'))?('T'):(((m_ReverseSequence == TRUE) && (a == 'T'))?('A'):(((m_ReverseSequence == TRUE) && (a == 'C'))?('G'):(((m_ReverseSequence == TRUE) && (a == 'G'))?('C'):(a)))))

#define FILEIO_MAXIMUM_HEADER_LENGTH_FOR_CLUSTERS     19

#define FILEIO_STRING_MAX                             128

/*********************************************************************************/
/* Private Type Definitions */
/*********************************************************************************/
typedef void (*FILEIO_POINTER_TO_STATE_FUNCTION) (BYTE);

typedef enum
{
   FILEIO_STATE_IDLE,
   FILEIO_STATE_HEADER_LINE,
   FILEIO_STATE_SEQUENCE,
} FILEIO_STATE;

typedef struct
{
   FILEIO_STATE State;
   FILEIO_POINTER_TO_STATE_FUNCTION StateFunction;
} FILEIO_STATE_ENTRY;

/*********************************************************************************/
/* Private Module Function Prototypes */
/*********************************************************************************/

/* States */
static void FILEIO_StateIdle (BYTE Ch);
static void FILEIO_StateHeaderLine(BYTE Ch);
static void FILEIO_StateSequence(BYTE Ch);

static void FILEIO_StoreSequencePart(void);

/*********************************************************************************/
/* Private Module Variables */
/*********************************************************************************/
static FILE *m_Infile;

static STRINGADT_STRING_TYPE *m_Sequence;
static STRINGADT_STRING_TYPE *m_SequenceReversed;
static STRINGADT_STRING_TYPE *m_HeaderLine;
static USHORT m_CurrentSequenceIndex;
static BOOLEAN m_ReverseSequence;
static BOOLEAN m_EndOfSequenceFound;
static BYTE m_SequencePart[FILEIO_SEQUENCE_PART_LENGTH];
static BYTE m_SequenceComplementPart[FILEIO_SEQUENCE_PART_LENGTH];

static FILEIO_STATE m_State;

const static FILEIO_STATE_ENTRY m_StateTable[] =
{
   {FILEIO_STATE_IDLE, FILEIO_StateIdle},
   {FILEIO_STATE_HEADER_LINE, FILEIO_StateHeaderLine},
   {FILEIO_STATE_SEQUENCE, FILEIO_StateSequence},
};

/*********************************************************************************/
/* Public and Private Module Functions */
/*********************************************************************************/

void FILEIO_OpenFASTASequenceFile(void)
{
   m_Infile = fopen(PARMS_GetInputFileName(), "r");
   if (m_Infile == NULL)
   {
      printf("Cannot open %s.\n", PARMS_GetInputFileName());
      exit(1);
   }
   m_State = FILEIO_STATE_IDLE;
}

/*********************************************************************************/

void FILEIO_CloseFASTASequenceFile(void)
{
   fclose(m_Infile);
}

/*********************************************************************************/

BOOLEAN FILEIO_ReadFASTASequences(STRINGADT_STRING_TYPE *Sequence,
                                 STRINGADT_STRING_TYPE *SequenceReversed,
                                 STRINGADT_STRING_TYPE *HeaderLine,
                                 BOOLEAN ReverseSequence)
{
   BYTE Ch;
   BOOLEAN FoundEndOfFile = FALSE;
   
   m_CurrentSequenceIndex = 0;
   m_Sequence = Sequence;
   m_SequenceReversed = SequenceReversed;
   m_HeaderLine = HeaderLine;
   m_ReverseSequence = ReverseSequence;
   m_EndOfSequenceFound = FALSE;

   if (m_Infile != NULL)
   {
      do
      {
         Ch = getc(m_Infile);
         m_StateTable[m_State].StateFunction(Ch);
      } while ((Ch != EOF) && (m_EndOfSequenceFound == FALSE));

      if (ReverseSequence == TRUE)
      {
         STRINGADT_ReverseString(SequenceReversed);
      }
      
      if (Ch == EOF)
      {
         FoundEndOfFile = TRUE;
      }
   }
   else
   {
      printf("File %s not open.\n", PARMS_GetInputFileName());
      exit(1);
   }
   
   return (FoundEndOfFile);
}

/*********************************************************************************/

void FILEIO_WriteClusters(CLUSTERLISTADT_CLUSTERLIST_TYPE *Clusters, 
                           STRINGLISTADT_STRINGLIST_TYPE *HeaderLines)
{
   FILE *outfile;
   FILE *repfile;
   ULONG i;
   ULONG j;
   BYTE Message[FILEIO_STRING_MAX];
   
   outfile = fopen(PARMS_GetOutputFileName(), "w");
   if (outfile != NULL)
   {
      repfile = fopen(PARMS_GetRepresentativeFileName(), "w");
      if (repfile != NULL)
      {   
         for (i = 0; i < Clusters->NumberOfClusters; i++)
         {
            fprintf(outfile, ">Cluster %lu\n", i);
            
            for (j = 0; j < Clusters->Clusters[i].NumberOfSequences; j++)
            {
               fprintf(outfile, "%lu\t>%s\n", j, HeaderLines->Strings[Clusters->Clusters[i].SequenceIndicies.Rows[0].Cols[j]].Characters);
            }

            fprintf(repfile, ">Cluster %lu, ", i);
            fprintf(repfile, ">%s\n", HeaderLines->Strings[Clusters->Clusters[i].SequenceIndicies.Rows[0].Cols[Clusters->Clusters[i].RepresentationIndex]].Characters);
   
            if (Clusters->Clusters[i].MarkedAsCompleteBasis == TRUE)
            {
               for (j = 0; j < (Clusters->Clusters[i].RepresentationSequence->Length - 1); j++)
               {
                  fprintf(repfile, "%c", Clusters->Clusters[i].RepresentationSequence->String[j]);
               }
            }
            else
            {
               for (j = 0; j < (Clusters->Clusters[i].Sequences.Trees[Clusters->Clusters[i].RepresentationIndex]->Length - 1); j++)
               {
                  fprintf(repfile, "%c", Clusters->Clusters[i].Sequences.Trees[Clusters->Clusters[i].RepresentationIndex]->String[j]);
               }
            }
            fprintf(repfile, "\n");

            sprintf (Message, "Progress: %lu / %lu.\r", i, Clusters->NumberOfClusters);
            PARMS_Printf(Message);
         }
         sprintf (Message, "Progress: %lu / %lu.\r\n", i, Clusters->NumberOfClusters);
         PARMS_Printf(Message);

         fclose(repfile);   
         fclose(outfile);
      }
      else
      {
         printf("Cannot create %s.\n", PARMS_GetRepresentativeFileName());
         exit(1);
      }
   }
   else
   {
      printf("Cannot create %s.\n", PARMS_GetOutputFileName());
      exit(1);
   }
}

/*********************************************************************************/

void FILEIO_StateIdle(BYTE Ch)
{
   if (Ch == '>')
   {
      m_State = FILEIO_STATE_HEADER_LINE;
   }
}

/*********************************************************************************/

void FILEIO_StateHeaderLine(BYTE Ch)
{
   if ((Ch == '\n') || (Ch == '\r'))
   {
      m_State = FILEIO_STATE_SEQUENCE;
      STRINGADT_AppendStringFromChar(m_HeaderLine, 
                                       '\0', 
                                       "FILEIO_StateHeaderLine");
   }
   else
   {
      if (m_HeaderLine->Length < FILEIO_MAXIMUM_HEADER_LENGTH_FOR_CLUSTERS)
      {
         if (Ch == ':')
         {
            STRINGADT_AppendStringFromChar(m_HeaderLine, 
                                             '_', 
                                             "FILEIO_StateHeaderLine");
         }
         else
         {
            STRINGADT_AppendStringFromChar(m_HeaderLine, 
                                             Ch, 
                                             "FILEIO_StateHeaderLine");
         }
      }
      else if (m_HeaderLine->Length == FILEIO_MAXIMUM_HEADER_LENGTH_FOR_CLUSTERS)
      {
         STRINGADT_AppendStringFromChars(m_HeaderLine, 
                                          (UBYTE *) "...", 
                                          3,
                                          "FILEIO_StateHeaderLine");
      }
   }
}

/*********************************************************************************/

void FILEIO_StateSequence(BYTE Ch)
{
   BYTE CH = toupper(Ch);
   
   switch(CH)
   {
      case '>':
         FILEIO_StoreSequencePart();
         m_EndOfSequenceFound = TRUE;
         m_State = FILEIO_STATE_HEADER_LINE;
         break;

      case 'A':
      case 'B':
      case 'C':
      case 'D':
      case 'E':
      case 'F':
      case 'G':
      case 'H':
      case 'I':
      case 'J':
      case 'K':
      case 'L':
      case 'M':
      case 'N':
      case 'O':
      case 'P':
      case 'Q':
      case 'R':
      case 'S':
      case 'T':
      case 'U':
      case 'V':
      case 'W':
      case 'X':
      case 'Y':
      case 'Z':
         m_SequencePart[m_CurrentSequenceIndex] = CH;
         m_SequenceComplementPart[m_CurrentSequenceIndex++] = FILEIO_BASE(CH);
         if (m_CurrentSequenceIndex >= FILEIO_SEQUENCE_PART_LENGTH)
         {
            FILEIO_StoreSequencePart();
         }
         break;

      case EOF:
         FILEIO_StoreSequencePart();
         m_EndOfSequenceFound = TRUE;
         break;

      default:
         break;
   }
}

/*********************************************************************************/

void FILEIO_StoreSequencePart(void)
{
   if (m_CurrentSequenceIndex > 0)
   {
      STRINGADT_AppendStringFromChars(m_Sequence, 
                                       (UBYTE *) m_SequencePart,
                                       m_CurrentSequenceIndex, 
                                       "FILEIO_StoreSequencePart");
      STRINGADT_AppendStringFromChars(m_SequenceReversed, 
                                       (UBYTE *) m_SequenceComplementPart,
                                       m_CurrentSequenceIndex, 
                                       "FILEIO_StoreSequencePart");
      m_CurrentSequenceIndex = 0;
   }
}
