/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  3-5-10                                                       */
/*                                                               */
/* 8/24/10  - speed up by only checking the first sequence in    */
/*             CheckAllBases instead of all basis sequences      */
/*             version 1.1                                       */
/* 9/24/10  - Added complete cluster searching, version 1.2      */
/* 9/27/10  - Added reverse complement checking, version 1.3     */
/*****************************************************************/

#define _GC

/*********************************************************************************/
/* Included Header Files */
/*********************************************************************************/
#include "Globals.h"
#include "GC.h"
#include "MatrixUlongADT.h"
#include "MatrixDoubleADT.h"
#include "StringADT.h"
#include "SuffixTreeADT.h"
#include "SuffixTreeListADT.h"
#include "StringListADT.h"
#include "Gram.h"
#include "ClusterADT.h"
#include "ClusterListADT.h"
#include "FileIO.h"
#include "Parameters.h"

/*********************************************************************************/
/* Private Module Constants */
/*********************************************************************************/
#define GC_MAX_STRING_LENGTH        128
#define GC_SUFFIX_TREE_TERMINATOR   '\0'
#define GC_MAX_DISTANCE             2.0

/*********************************************************************************/
/* Private Type Definitions */
/*********************************************************************************/

/*********************************************************************************/
/* Private Module Function Prototypes */
/*********************************************************************************/
static void GC_AddSequenceToCluster(CLUSTERLISTADT_CLUSTERLIST_TYPE *Clusters, 
                                    STADT_TREE_TYPE *Sequence,
                                    STADT_TREE_TYPE *SequenceReversed,
                                    ULONG SequenceIndex,
                                    BOOLEAN *UsedReverseSequence);
static BOOLEAN GC_CheckAllBases(CLUSTERADT_CLUSTER_TYPE *Cluster, 
                                 STADT_TREE_TYPE *Sequence,
                                 STADT_TREE_TYPE *SequenceReversed,
                                 ULONG SequenceIndex,
                                 FLOAT DictionarySize,
                                 FLOAT DictionaryReversedSize,
                                 FLOAT GrammarDistanceMaxThreshold,
                                 FLOAT *MinDistance,
                                 BOOLEAN AddIfFound,
                                 BOOLEAN *UsedReverseSequence);
static BOOLEAN GC_CheckCentroid(CLUSTERADT_CLUSTER_TYPE *Cluster, 
                                 STADT_TREE_TYPE *Sequence,
                                 STADT_TREE_TYPE *SequenceReversed,
                                 ULONG SequenceIndex,
                                 FLOAT DictionarySize,
                                 FLOAT DictionaryReversedSize,
                                 FLOAT GrammarDistanceCentroidMaxThreshold,
                                 FLOAT *MinDistance,
                                 BOOLEAN AddIfFound,
                                 BOOLEAN *UsedReverseSequence);
static void GC_CalculateCentroid(CLUSTERADT_CLUSTER_TYPE *Cluster);

/*********************************************************************************/
/* Private Module Variables */
/*********************************************************************************/

/*********************************************************************************/
/* Public and Private Module Functions */
/*********************************************************************************/

INT main(INT argc, BYTE **argv)
{
   BYTE Message[GC_MAX_STRING_LENGTH];
   STADT_TREE_TYPE *Tree;
   STADT_TREE_TYPE *TreeReversed;
   STRINGADT_STRING_TYPE Sequence;
   STRINGADT_STRING_TYPE SequenceReversed;
   STRINGADT_STRING_TYPE HeaderLine;
   STRINGLISTADT_STRINGLIST_TYPE HeaderLines;
   CLUSTERLISTADT_CLUSTERLIST_TYPE Clusters;
   BOOLEAN FoundEndOfFile = FALSE;
   ULONG SequenceCount;
   BOOLEAN UsedReverseSequence;
   
   if (PARMS_ProcessCommandLine(argc, argv) == TRUE)
   {
      CLUSTERLISTADT_InitializeClusterList(&Clusters);
      STRINGLISTADT_InitializeStringList(&HeaderLines);

      srand((int) time(NULL));
      
      FILEIO_OpenFASTASequenceFile();
      SequenceCount = 0;
      while (FoundEndOfFile == FALSE)
      {
         STRINGADT_InitializeString(&Sequence);
         STRINGADT_InitializeString(&SequenceReversed);
         STRINGADT_InitializeString(&HeaderLine);

         FoundEndOfFile = FILEIO_ReadFASTASequences(&Sequence, &SequenceReversed, &HeaderLine, TRUE);

         STRINGLISTADT_AppendStringToStringList(&HeaderLines, &HeaderLine);         

         Tree = (STADT_TREE_TYPE *) malloc (sizeof (STADT_TREE_TYPE));
         if (Tree == NULL)
         {
            printf("Error: malloc failed in main.\n");
            exit(1);
         }

         STADT_CreateTree (Tree, Sequence.Characters, Sequence.Length, GC_SUFFIX_TREE_TERMINATOR);

         TreeReversed = (STADT_TREE_TYPE *) malloc (sizeof (STADT_TREE_TYPE));
         if (TreeReversed == NULL)
         {
            printf("Error: malloc failed in main.\n");
            exit(1);
         }

         STADT_CreateTree (TreeReversed, SequenceReversed.Characters, SequenceReversed.Length, GC_SUFFIX_TREE_TERMINATOR);

         GC_AddSequenceToCluster(&Clusters, Tree, TreeReversed, SequenceCount, &UsedReverseSequence);
         
         if (UsedReverseSequence == FALSE)
         {
            STADT_DeleteTree(TreeReversed);
            free(TreeReversed);
         }
         else
         {
            STADT_DeleteTree(Tree);
            free(Tree);
         }
         
         sprintf (Message, "Progress: %lu sequences, %lu clusters.\r", SequenceCount, Clusters.NumberOfClusters);
         PARMS_Printf(Message);

         STRINGADT_CleanUpString(&Sequence);
         STRINGADT_CleanUpString(&SequenceReversed);
         STRINGADT_CleanUpString(&HeaderLine);
         
         SequenceCount++;   
      }
      FILEIO_CloseFASTASequenceFile();

      sprintf (Message, "Progress: %lu sequences, %lu clusters.\r\n", SequenceCount, Clusters.NumberOfClusters);
      PARMS_Printf(Message);

      PARMS_Printf("Writing cluster file.\n");
      FILEIO_WriteClusters(&Clusters, &HeaderLines);

/*
      CLUSTERLISTADT_CleanUpClusterList(&Clusters);
      STRINGLISTADT_CleanUpStringList(&HeaderLines);
*/
      PARMS_Printf("Done.\n");
   }
   
   return(0);
}

/*********************************************************************************/

void GC_AddSequenceToCluster(CLUSTERLISTADT_CLUSTERLIST_TYPE *Clusters,
                              STADT_TREE_TYPE *Sequence,
                              STADT_TREE_TYPE *SequenceReversed,
                              ULONG SequenceIndex,
                              BOOLEAN *UsedReverseSequence)
{
   ULONG i = 0;
   BOOLEAN Found = FALSE;
   FLOAT DictionarySize;
   FLOAT DictionaryReversedSize;
   FLOAT GrammarDistanceMaxThreshold;
   FLOAT GrammarDistanceCentroidMaxThreshold;
   FLOAT Distance;
   FLOAT MinDistance = GC_MAX_DISTANCE;
   ULONG MinIndex = 0;
   BOOLEAN MinSet = FALSE;   

   GrammarDistanceMaxThreshold = PARMS_GetGrammarDistanceMaxThreshold();
   GrammarDistanceCentroidMaxThreshold = PARMS_GetGrammarDistanceCentroidMaxThreshold();

   DictionarySize = GRAM_CreateIndividualDictionary(Sequence);
   DictionaryReversedSize = GRAM_CreateIndividualDictionary(SequenceReversed);
   
   if (PARMS_GetCompleteClusterSearch() == TRUE)
   {
      while (i < Clusters->NumberOfClusters)
      {
         if (Clusters->Clusters[i].NumberOfBases < PARMS_GetFullBasisAmount())
         {
            Found = GC_CheckAllBases(&(Clusters->Clusters[i]), 
                                       Sequence, 
                                       SequenceReversed,
                                       SequenceIndex, 
                                       DictionarySize,
                                       DictionaryReversedSize,
                                       GrammarDistanceMaxThreshold,
                                       &Distance,
                                       FALSE,
                                       UsedReverseSequence);
         }
         else
         {
            Found = GC_CheckCentroid(&(Clusters->Clusters[i]), 
                                       Sequence, 
                                       SequenceReversed,
                                       SequenceIndex, 
                                       DictionarySize,
                                       DictionaryReversedSize,
                                       GrammarDistanceCentroidMaxThreshold,
                                       &Distance,
                                       FALSE,
                                       UsedReverseSequence);
         }
         
         if (Found == TRUE)
         {
            if (MinSet == FALSE)
            {
               MinSet = TRUE;
               MinDistance = Distance;
               MinIndex = i;
            }
            else
            {
               if (Distance < MinDistance)
               {
                  MinDistance = Distance;
                  MinIndex = i;
               }
            }
         }

         i++;
      }
      
      if (MinSet == TRUE)
      {
         if (Clusters->Clusters[MinIndex].NumberOfBases < PARMS_GetFullBasisAmount())
         {
            Found = GC_CheckAllBases(&(Clusters->Clusters[MinIndex]), 
                                       Sequence, 
                                       SequenceReversed,
                                       SequenceIndex, 
                                       DictionarySize,
                                       DictionaryReversedSize,
                                       GrammarDistanceMaxThreshold,
                                       &Distance,
                                       TRUE,
                                       UsedReverseSequence);
   
            if (Clusters->Clusters[MinIndex].NumberOfBases >= PARMS_GetFullBasisAmount())
            {
               GC_CalculateCentroid(&(Clusters->Clusters[MinIndex]));
               CLUSTERADT_SetClusterAsCompleteBasis(&(Clusters->Clusters[MinIndex]));
            }
         }
         else
         {
            Found = GC_CheckCentroid(&(Clusters->Clusters[MinIndex]), 
                                       Sequence, 
                                       SequenceReversed,
                                       SequenceIndex, 
                                       DictionarySize,
                                       DictionaryReversedSize,
                                       GrammarDistanceCentroidMaxThreshold,
                                       &Distance,
                                       TRUE,
                                       UsedReverseSequence);
         }
      }      
   }
   else
   {
      while ((Found == FALSE) && (i < Clusters->NumberOfClusters))
      {
         if (Clusters->Clusters[i].NumberOfBases < PARMS_GetFullBasisAmount())
         {
            Found = GC_CheckAllBases(&(Clusters->Clusters[i]), 
                                       Sequence, 
                                       SequenceReversed,
                                       SequenceIndex, 
                                       DictionarySize,
                                       DictionaryReversedSize,
                                       GrammarDistanceMaxThreshold,
                                       &Distance,
                                       TRUE,
                                       UsedReverseSequence);
   
            if (Clusters->Clusters[i].NumberOfBases >= PARMS_GetFullBasisAmount())
            {
               GC_CalculateCentroid(&(Clusters->Clusters[i]));
               CLUSTERADT_SetClusterAsCompleteBasis(&(Clusters->Clusters[i]));
            }
         }
         else
         {
            Found = GC_CheckCentroid(&(Clusters->Clusters[i]), 
                                       Sequence, 
                                       SequenceReversed,
                                       SequenceIndex, 
                                       DictionarySize,
                                       DictionaryReversedSize,
                                       GrammarDistanceCentroidMaxThreshold,
                                       &Distance,
                                       TRUE,
                                       UsedReverseSequence);
         }
         
         if (Found == FALSE)
         {
            i++;
         }
      }
   }
      
   if (Found == FALSE)
   {
      CLUSTERLISTADT_AppendNewClusterToClusterList(Clusters, Sequence, SequenceIndex, DictionarySize, PARMS_GetFullBasisAmount());

      *UsedReverseSequence = FALSE;

      if (PARMS_GetFullBasisAmount() == 1)
      {
         CLUSTERADT_SetClusterAsCompleteBasis(&(Clusters->Clusters[Clusters->NumberOfClusters - 1]));
      }
   }
}

/*********************************************************************************/

BOOLEAN GC_CheckAllBases(CLUSTERADT_CLUSTER_TYPE *Cluster, 
                           STADT_TREE_TYPE *Sequence,
                           STADT_TREE_TYPE *SequenceReversed,
                           ULONG SequenceIndex,
                           FLOAT DictionarySize,
                           FLOAT DictionaryReversedSize,
                           FLOAT GrammarDistanceMaxThreshold,
                           FLOAT *MinDistance,
                           BOOLEAN AddIfFound,
                           BOOLEAN *UsedReverseSequence)
{
   BOOLEAN Found = FALSE;
   ULONG j;
   FLOAT Distance;
   FLOAT DistanceReversed = GC_MAX_DISTANCE;
   ULONG MinIndex;

   Distance = GRAM_CalculateDistance(Cluster->Sequences.Trees[Cluster->BasisSequenceIndicies.Rows[0].Cols[0]], 
                                       Cluster->SequenceDictionaries.Rows[0].Cols[Cluster->BasisSequenceIndicies.Rows[0].Cols[0]],
                                       Sequence,
                                       DictionarySize);

   if (PARMS_GetCheckReverseComplements() == TRUE)
   {
      DistanceReversed = GRAM_CalculateDistance(Cluster->Sequences.Trees[Cluster->BasisSequenceIndicies.Rows[0].Cols[0]], 
                                          Cluster->SequenceDictionaries.Rows[0].Cols[Cluster->BasisSequenceIndicies.Rows[0].Cols[0]],
                                          SequenceReversed,
                                          DictionaryReversedSize);
   }
   
   if ((Distance < DistanceReversed) || (PARMS_GetCheckReverseComplements() == FALSE))
   {
      *MinDistance = GC_MAX_DISTANCE;
      if (Distance < GrammarDistanceMaxThreshold)
      {
         Found = TRUE;
   
         MinIndex = 0;
         *MinDistance = Distance;
         for (j = 1; j < Cluster->NumberOfBases; j++)
         {
            Distance = GRAM_CalculateDistance(Cluster->Sequences.Trees[Cluster->BasisSequenceIndicies.Rows[0].Cols[j]], 
                                                Cluster->SequenceDictionaries.Rows[0].Cols[Cluster->BasisSequenceIndicies.Rows[0].Cols[j]],
                                                Sequence,
                                                DictionarySize);
      
            if (Distance < *MinDistance)
            {
               MinIndex = j;
               *MinDistance = Distance;
            }
         }
   
         if (AddIfFound == TRUE)
         {
            if (*MinDistance >= PARMS_GetGrammarDistanceIdenticalThreshold())
            {
               CLUSTERADT_AddSequenceIncompleteBasis(Cluster, Sequence, SequenceIndex, DictionarySize, TRUE);
            }
            else
            {
               CLUSTERADT_AddSequenceIncompleteBasis(Cluster, Sequence, SequenceIndex, DictionarySize, FALSE);
            }

            *UsedReverseSequence = FALSE;
         }
      }
   }
   else
   {
      *MinDistance = GC_MAX_DISTANCE;
      if (DistanceReversed < GrammarDistanceMaxThreshold)
      {
         Found = TRUE;
   
         MinIndex = 0;
         *MinDistance = DistanceReversed;
         for (j = 1; j < Cluster->NumberOfBases; j++)
         {
            DistanceReversed = GRAM_CalculateDistance(Cluster->Sequences.Trees[Cluster->BasisSequenceIndicies.Rows[0].Cols[j]], 
                                                Cluster->SequenceDictionaries.Rows[0].Cols[Cluster->BasisSequenceIndicies.Rows[0].Cols[j]],
                                                SequenceReversed,
                                                DictionaryReversedSize);
      
            if (DistanceReversed < *MinDistance)
            {
               MinIndex = j;
               *MinDistance = DistanceReversed;
            }
         }
   
         if (AddIfFound == TRUE)
         {
            if (*MinDistance >= PARMS_GetGrammarDistanceIdenticalThreshold())
            {
               CLUSTERADT_AddSequenceIncompleteBasis(Cluster, SequenceReversed, SequenceIndex, DictionaryReversedSize, TRUE);
            }
            else
            {
               CLUSTERADT_AddSequenceIncompleteBasis(Cluster, SequenceReversed, SequenceIndex, DictionaryReversedSize, FALSE);
            }

            *UsedReverseSequence = TRUE;
         }
      }
   }

   return (Found);
}

/*********************************************************************************/

BOOLEAN GC_CheckCentroid(CLUSTERADT_CLUSTER_TYPE *Cluster, 
                           STADT_TREE_TYPE *Sequence,
                           STADT_TREE_TYPE *SequenceReversed,
                           ULONG SequenceIndex,
                           FLOAT DictionarySize,
                           FLOAT DictionaryReversedSize,
                           FLOAT GrammarDistanceCentroidMaxThreshold,
                           FLOAT *MinDistance,
                           BOOLEAN AddIfFound,
                           BOOLEAN *UsedReverseSequence)
{
   BOOLEAN Found = FALSE;
   FLOAT Distance;
   FLOAT DistanceReversed = GC_MAX_DISTANCE;
   
   Distance = GRAM_CalculateDistance(Cluster->RepresentationSequence, 
                                    Cluster->RepresentationDictionary,
                                    Sequence,
                                    DictionarySize);

   if (PARMS_GetCheckReverseComplements() == TRUE)
   {
      DistanceReversed = GRAM_CalculateDistance(Cluster->RepresentationSequence, 
                                       Cluster->RepresentationDictionary,
                                       SequenceReversed,
                                       DictionaryReversedSize);
   }
   
   if ((Distance < DistanceReversed) || (PARMS_GetCheckReverseComplements() == FALSE))
   {
      *MinDistance = GC_MAX_DISTANCE;
      if (Distance < GrammarDistanceCentroidMaxThreshold)
      {
         Found = TRUE;
         *MinDistance = Distance;
         if (AddIfFound == TRUE)
         {
            CLUSTERADT_AddSequenceCompleteBasis(Cluster, Sequence, SequenceIndex, DictionarySize);

            *UsedReverseSequence = FALSE;
         }
      }
   }
   else
   {
      *MinDistance = GC_MAX_DISTANCE;
      if (DistanceReversed < GrammarDistanceCentroidMaxThreshold)
      {
         Found = TRUE;
         *MinDistance = DistanceReversed;
         if (AddIfFound == TRUE)
         {
            CLUSTERADT_AddSequenceCompleteBasis(Cluster, SequenceReversed, SequenceIndex, DictionaryReversedSize);

            *UsedReverseSequence = TRUE;
         }
      }
   }
         
   return (Found);
}

/*********************************************************************************/

void GC_CalculateCentroid(CLUSTERADT_CLUSTER_TYPE *Cluster)
{
   ULONG i;
   ULONG j;
   DOUBLE MinDistance = GC_MAX_DISTANCE;
   BOOLEAN MinSet;
   DOUBLE Temp;
   MATRIXDOUBLEADT_MATRIX_TYPE CentroidVector;

   MATRIXDOUBLEADT_InitializeMatrix(&CentroidVector, 1, Cluster->NumberOfBases);
      
   for (i = 0; i < Cluster->NumberOfSequences; i++)
   {
      if (i == 0)
      {
         for (j = 0; j < Cluster->NumberOfBases; j++)
         {
            CentroidVector.Rows[0].Cols[j] = Cluster->DistanceVectors.Rows[i].Cols[j];
         }
      }
      else
      {
         for (j = 0; j < Cluster->NumberOfBases; j++)
         {
            CentroidVector.Rows[0].Cols[j] += Cluster->DistanceVectors.Rows[i].Cols[j];
         }
      }
   }

   for (j = 0; j < Cluster->NumberOfBases; j++)
   {
      CentroidVector.Rows[0].Cols[j] /= ((DOUBLE) Cluster->NumberOfSequences);
   }

   MinSet = FALSE;
   for (i = 0; i < Cluster->NumberOfSequences; i++)
   {
      Temp = 0;
      for (j = 0; j < Cluster->NumberOfBases; j++)
      {
         Temp += ((Cluster->DistanceVectors.Rows[i].Cols[j] - CentroidVector.Rows[0].Cols[j]) * 
                  (Cluster->DistanceVectors.Rows[i].Cols[j] - CentroidVector.Rows[0].Cols[j]));
      }
         
      if (MinSet == FALSE)
      {
         MinSet = TRUE;
         Cluster->RepresentationIndex = i;
         MinDistance = Temp;
      }
      else
      {
         if (MinDistance > Temp)
         {
            Cluster->RepresentationIndex = i;
            MinDistance = Temp;
         }
         else if (MinDistance == Temp)
         {
            if (COIN_FLIP)
            {
               Cluster->RepresentationIndex = i;
               MinDistance = Temp;
            }
         }
      }
   }

   MATRIXDOUBLEADT_CleanUpMatrix(&CentroidVector);
}
