/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  3-5-10                                                       */
/*                                                               */
/* 8/24/10  - Removed the short sequence parameters, version 1.1 */
/* 9/24/10  - Added complete cluster searching, version 1.2      */
/* 9/27/10  - Added reverse complement checking, version 1.3     */
/*****************************************************************/

#define _PARMS

/*********************************************************************************/
/* Included Header Files */
/*********************************************************************************/
#include "Globals.h"
#include "Parameters.h"

/*********************************************************************************/
/* Private Module Constants */
/*********************************************************************************/
#define PARMS_STARTING_ARGUMENT_POS                                  2
#define PARMS_DELIMITER_POS                                          0
#define PARMS_DELIMITER_CHAR                                         '-'
#define PARMS_OPTION_POS                                             1

#define PARMS_MAX_ARGLENGTH                                          1024
#define PARMS_MAX_STRING_LENGTH                                      1024

#define PARMS_DEFAULT_INFILE                                         "infile"
#define PARMS_DEFAULT_OUTFILE                                        "outfile"
#define PARMS_DEFAULT_REPRESENTATIVEFILE                             "repfile"

#define PARMS_DEFAULT_FULL_BASIS_AMOUNT                              4
#define PARMS_DEFAULT_GRAMMAR_DISTANCE_IDENTICAL_THRESHOLD           0.01
#define PARMS_DEFAULT_GRAMMAR_DISTANCE_MAX_THRESHOLD                 0.13
#define PARMS_DEFAULT_GRAMMAR_DISTANCE_CENTROID_MAX_THRESHOLD        0.13

#define PARMS_DEFAULT_QUIET_MODE                                     FALSE

/*********************************************************************************/
/* Private Type Definitions */
/*********************************************************************************/

/*********************************************************************************/
/* Private Module Variables */
/*********************************************************************************/
static BYTE m_InputFileName[PARMS_MAX_STRING_LENGTH];
static BYTE m_OutputFileName[PARMS_MAX_STRING_LENGTH];
static BYTE m_RepresentativeFileName[PARMS_MAX_STRING_LENGTH];

static UBYTE m_FullBasisAmount;
static FLOAT m_GrammarDistanceMaxThreshold;
static FLOAT m_GrammarDistanceIdenticalThreshold;
static FLOAT m_GrammarDistanceCentroidMaxThreshold;
static BOOLEAN m_CompleteClusterSearch;
static BOOLEAN m_CheckReverseComplements;

static BOOLEAN m_QuietMode;

/*********************************************************************************/
/* Private Module Function Prototypes */
/*********************************************************************************/
static BYTE PARMS_GetNextOption(INT argc, 
                                 BYTE **argv, 
                                 BYTE *OptionDelimiters, 
                                 BYTE *OptionArgument, 
                                 INT *CurrentOptionIndex);
static void PARMS_InitializeDefaultValues(void);
static void PARMS_OutputHeaderText(void);
static void PARMS_OutputHelpText(void);
                           
/*********************************************************************************/
/* Public and Private Module Functions */
/*********************************************************************************/

BYTE *PARMS_GetInputFileName(void)
{
   return (m_InputFileName);
}

/*********************************************************************************/

BYTE *PARMS_GetOutputFileName(void)
{
   return (m_OutputFileName);
}

/*********************************************************************************/

BYTE *PARMS_GetRepresentativeFileName(void)
{
   return (m_RepresentativeFileName);
}

/*********************************************************************************/

UBYTE PARMS_GetFullBasisAmount(void)
{
   return (m_FullBasisAmount);
}

/*********************************************************************************/

FLOAT PARMS_GetGrammarDistanceIdenticalThreshold(void)
{
   return (m_GrammarDistanceIdenticalThreshold);
}

/*********************************************************************************/

FLOAT PARMS_GetGrammarDistanceMaxThreshold(void)
{
   return (m_GrammarDistanceMaxThreshold);
}

/*********************************************************************************/

FLOAT PARMS_GetGrammarDistanceCentroidMaxThreshold(void)
{
   return (m_GrammarDistanceCentroidMaxThreshold);
}

/*********************************************************************************/

BOOLEAN PARMS_GetCompleteClusterSearch(void)
{
   return (m_CompleteClusterSearch);
}

/*********************************************************************************/

BOOLEAN PARMS_GetCheckReverseComplements(void)
{
   return (m_CheckReverseComplements);
}

/*********************************************************************************/

BOOLEAN PARMS_ProcessCommandLine(INT argc, BYTE **argv)
{
   BYTE Option;
   BYTE OptionArgument[PARMS_MAX_ARGLENGTH];
   INT CurrentOption = PARMS_STARTING_ARGUMENT_POS;
   BOOLEAN OkToRun = TRUE;
   
   PARMS_InitializeDefaultValues();
   PARMS_OutputHeaderText();
   
   while ((Option = PARMS_GetNextOption(argc, argv, "i:o:r:B:C:G:b", OptionArgument, &CurrentOption)) != EOF)
   {
      switch(Option)
      {
         case 'i':
            if (strlen(OptionArgument) > 0)
            {
               sprintf(m_InputFileName, "%s", OptionArgument);
            }
            else
            {
               printf("No input file specified.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
            break;

         case 'o':
            if (strlen(OptionArgument) > 0)
            {
               sprintf(m_OutputFileName, "%s", OptionArgument);
            }
            else
            {
               printf("No output file specified.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
            break;

         case 'r':
            if (strlen(OptionArgument) > 0)
            {
               sprintf(m_RepresentativeFileName, "%s", OptionArgument);
            }
            else
            {
               printf("No representative file specified.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
            break;

         case 'h':
            PARMS_OutputHelpText();
            OkToRun = FALSE;
            break;
            
         case 'q':
            m_QuietMode = TRUE;
            break;

         case 'B':
            if (strlen(OptionArgument) > 0)
            {
               m_FullBasisAmount = atoi(OptionArgument);
               if (m_FullBasisAmount < 1)
               {
                  printf("Invalid -B option value.\n\n");
                  OkToRun = FALSE;
               }
            }
            else
            {
               printf("Unknown option value.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
            break;
               
         case 'C':
            if (strlen(OptionArgument) > 0)
            {
               m_GrammarDistanceCentroidMaxThreshold = atof(OptionArgument);
            }
            else
            {
               printf("Unknown option value.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
            break;
               
         case 'G':
            if (strlen(OptionArgument) > 0)
            {
               m_GrammarDistanceMaxThreshold = atof(OptionArgument);
            }
            else
            {
               printf("Unknown option value.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
            break;
               
         case 'b':
            if (strlen(OptionArgument) > 0)
            {
               m_GrammarDistanceIdenticalThreshold = atof(OptionArgument);
            }
            else
            {
               printf("Unknown option value.  Use -h for a list of options.\n\n");
               OkToRun = FALSE;
            }
            break;
               
         case 'c':
            m_CompleteClusterSearch = TRUE;
            break;

         case 'R':
            m_CheckReverseComplements = TRUE;
            break;

         default:
            printf("Unknown option.  Use -h for a list of options.\n\n");
            OkToRun = FALSE;
            break;
      }
   }

   return (OkToRun);
}

/*********************************************************************************/

void PARMS_Printf(BYTE *String)
{
   if (m_QuietMode == FALSE)
   {
      fprintf(stderr, "%s", String);
   }
}

/*********************************************************************************/

BYTE PARMS_GetNextOption(INT argc, 
                           BYTE **argv, 
                           BYTE *OptionDelimiters, 
                           BYTE *OptionArgument, 
                           INT *CurrentOptionIndex)
{
   BYTE ReturnOption = EOF;
   BOOLEAN Done;

   OptionArgument[0] = 0;

   if (argc >= PARMS_STARTING_ARGUMENT_POS)
   {
      if (*CurrentOptionIndex <= argc)
      {
         if (*CurrentOptionIndex < PARMS_STARTING_ARGUMENT_POS)
         {
            *CurrentOptionIndex = PARMS_STARTING_ARGUMENT_POS;
         }

         Done = FALSE;
         while (Done == FALSE)
         {
            if (*CurrentOptionIndex <= argc)
            {
               if (argv[*CurrentOptionIndex - 1][PARMS_DELIMITER_POS] != PARMS_DELIMITER_CHAR)
               {
                  *CurrentOptionIndex += 1;
               }
               else
               {
                  Done = TRUE;
               }
            }
            else
            {
               Done = TRUE;
            }
         }

         if (*CurrentOptionIndex <= argc)
         {
            ReturnOption = argv[*CurrentOptionIndex - 1][PARMS_OPTION_POS];
            *CurrentOptionIndex += 1;

            if (*CurrentOptionIndex <= argc)
            {
               if (argv[*CurrentOptionIndex - 1][PARMS_DELIMITER_POS] != PARMS_DELIMITER_CHAR)
               {
                  if (strchr(OptionDelimiters, ReturnOption) != NULL)
                  {
                     strcpy(OptionArgument, argv[*CurrentOptionIndex - 1]);
                     *CurrentOptionIndex += 1;
                  }
               }
            }
         }
      }
   }

   return (ReturnOption);
}

/*********************************************************************************/

void PARMS_InitializeDefaultValues(void)
{
   sprintf(m_InputFileName, "%s", PARMS_DEFAULT_INFILE);
   sprintf(m_OutputFileName, "%s", PARMS_DEFAULT_OUTFILE);
   sprintf(m_RepresentativeFileName, "%s", PARMS_DEFAULT_REPRESENTATIVEFILE);

   m_FullBasisAmount = PARMS_DEFAULT_FULL_BASIS_AMOUNT;
   m_GrammarDistanceMaxThreshold = PARMS_DEFAULT_GRAMMAR_DISTANCE_MAX_THRESHOLD;
   m_GrammarDistanceIdenticalThreshold = PARMS_DEFAULT_GRAMMAR_DISTANCE_IDENTICAL_THRESHOLD;
   m_GrammarDistanceCentroidMaxThreshold = PARMS_DEFAULT_GRAMMAR_DISTANCE_CENTROID_MAX_THRESHOLD;
   m_CompleteClusterSearch = FALSE;
   m_CheckReverseComplements = FALSE;

   m_QuietMode = PARMS_DEFAULT_QUIET_MODE;
}

/*********************************************************************************/

void PARMS_OutputHeaderText(void)
{
   if (m_QuietMode == FALSE)
   {
      printf("\n   GramCluster -- ");
      printf("Version %d.%d -- ", VERSION_MAJOR, VERSION_MINOR);
      printf("%s\n", VERSION_DATE);
      printf("   University of Nebraska-Lincoln\n");
      printf("   Department of Electrical Engineering\n");
      printf("   David Russell\n");
      printf("\n");
   }
}

/*********************************************************************************/

void PARMS_OutputHelpText(void)
{
   printf("\tUsage: GramCluster [options]\n\n");
   printf("\tOptions:\n\n");
   printf("\t-i filename\tSpecify input FASTA file (default is '%s')\n", PARMS_DEFAULT_INFILE);
   printf("\t-o filename\tSpecify output file (default is '%s')\n", PARMS_DEFAULT_OUTFILE);
   printf("\t-r filename\tSpecify representative file (default is '%s')\n", PARMS_DEFAULT_REPRESENTATIVEFILE);
   printf("\t-h \t\tHelp\n");
   printf("\t-q \t\tQuiet mode (default is verbose)\n");
   printf("\n");
   printf("\t-B value\tSpecify the full basis amount (default is %d)\n", PARMS_DEFAULT_FULL_BASIS_AMOUNT);
   printf("\t-b value\tSpecify the grammar distance identical threshold (default is %2.4f)\n", PARMS_DEFAULT_GRAMMAR_DISTANCE_IDENTICAL_THRESHOLD);
   printf("\t-C value\tSpecify the grammar distance to centroid max threshold (default is %2.4f)\n", PARMS_DEFAULT_GRAMMAR_DISTANCE_CENTROID_MAX_THRESHOLD);
   printf("\t-G value\tSpecify the grammar distance max threshold (default is %2.4f)\n", PARMS_DEFAULT_GRAMMAR_DISTANCE_MAX_THRESHOLD);
   printf("\t-c \t\tPerform complete cluster searching (default is greedy searching)\n");
   printf("\t-R \t\tCheck reverse complements also (default is no reverse complements)\n");
   printf("\n\n");
}
